# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'boris/observation.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(959, 677)
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.lb_star = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.lb_star.setFont(font)
        self.lb_star.setStyleSheet("color: red")
        self.lb_star.setObjectName("lb_star")
        self.horizontalLayout_2.addWidget(self.lb_star)
        self.leObservationId = QtWidgets.QLineEdit(Form)
        self.leObservationId.setObjectName("leObservationId")
        self.horizontalLayout_2.addWidget(self.leObservationId)
        self.label_8 = QtWidgets.QLabel(Form)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_2.addWidget(self.label_8)
        self.dteDate = QtWidgets.QDateTimeEdit(Form)
        self.dteDate.setCalendarPopup(True)
        self.dteDate.setObjectName("dteDate")
        self.horizontalLayout_2.addWidget(self.dteDate)
        self.verticalLayout_6.addLayout(self.horizontalLayout_2)
        self.splitter = QtWidgets.QSplitter(Form)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_9 = QtWidgets.QLabel(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_9.sizePolicy().hasHeightForWidth())
        self.label_9.setSizePolicy(sizePolicy)
        self.label_9.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label_9.setObjectName("label_9")
        self.verticalLayout_2.addWidget(self.label_9)
        self.teDescription = QtWidgets.QTextEdit(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.teDescription.sizePolicy().hasHeightForWidth())
        self.teDescription.setSizePolicy(sizePolicy)
        self.teDescription.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.teDescription.setAcceptDrops(False)
        self.teDescription.setObjectName("teDescription")
        self.verticalLayout_2.addWidget(self.teDescription)
        self.cb_time_offset = QtWidgets.QCheckBox(self.layoutWidget)
        self.cb_time_offset.setObjectName("cb_time_offset")
        self.verticalLayout_2.addWidget(self.cb_time_offset)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.lbTimeOffset = QtWidgets.QLabel(self.layoutWidget)
        self.lbTimeOffset.setObjectName("lbTimeOffset")
        self.horizontalLayout_6.addWidget(self.lbTimeOffset)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem)
        self.verticalLayout_2.addLayout(self.horizontalLayout_6)
        self.cb_observation_time_interval = QtWidgets.QCheckBox(self.layoutWidget)
        self.cb_observation_time_interval.setObjectName("cb_observation_time_interval")
        self.verticalLayout_2.addWidget(self.cb_observation_time_interval)
        self.layoutWidget1 = QtWidgets.QWidget(self.splitter)
        self.layoutWidget1.setObjectName("layoutWidget1")
        self.verticalLayout_11 = QtWidgets.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_11.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_11.setObjectName("verticalLayout_11")
        self.label_3 = QtWidgets.QLabel(self.layoutWidget1)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_11.addWidget(self.label_3)
        self.twIndepVariables = QtWidgets.QTableWidget(self.layoutWidget1)
        self.twIndepVariables.setObjectName("twIndepVariables")
        self.twIndepVariables.setColumnCount(3)
        self.twIndepVariables.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.twIndepVariables.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.twIndepVariables.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.twIndepVariables.setHorizontalHeaderItem(2, item)
        self.verticalLayout_11.addWidget(self.twIndepVariables)
        self.verticalLayout_6.addWidget(self.splitter)
        self.gb_observation_type = QtWidgets.QGroupBox(Form)
        self.gb_observation_type.setObjectName("gb_observation_type")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout(self.gb_observation_type)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.rb_media_files = QtWidgets.QRadioButton(self.gb_observation_type)
        self.rb_media_files.setObjectName("rb_media_files")
        self.horizontalLayout_4.addWidget(self.rb_media_files)
        self.rb_live = QtWidgets.QRadioButton(self.gb_observation_type)
        self.rb_live.setObjectName("rb_live")
        self.horizontalLayout_4.addWidget(self.rb_live)
        self.rb_images = QtWidgets.QRadioButton(self.gb_observation_type)
        self.rb_images.setEnabled(True)
        self.rb_images.setObjectName("rb_images")
        self.horizontalLayout_4.addWidget(self.rb_images)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.horizontalLayout_7.addLayout(self.horizontalLayout_4)
        self.verticalLayout_6.addWidget(self.gb_observation_type)
        self.sw_observation_type = QtWidgets.QStackedWidget(Form)
        self.sw_observation_type.setObjectName("sw_observation_type")
        self.page = QtWidgets.QWidget()
        self.page.setObjectName("page")
        self.sw_observation_type.addWidget(self.page)
        self.pg_media_files = QtWidgets.QWidget()
        self.pg_media_files.setObjectName("pg_media_files")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.pg_media_files)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.tabWidget = QtWidgets.QTabWidget(self.pg_media_files)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_player_1 = QtWidgets.QWidget()
        self.tab_player_1.setObjectName("tab_player_1")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.tab_player_1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.twVideo1 = QtWidgets.QTableWidget(self.tab_player_1)
        self.twVideo1.setEditTriggers(QtWidgets.QAbstractItemView.AnyKeyPressed|QtWidgets.QAbstractItemView.DoubleClicked|QtWidgets.QAbstractItemView.EditKeyPressed)
        self.twVideo1.setAlternatingRowColors(True)
        self.twVideo1.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.twVideo1.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.twVideo1.setTextElideMode(QtCore.Qt.ElideNone)
        self.twVideo1.setObjectName("twVideo1")
        self.twVideo1.setColumnCount(7)
        self.twVideo1.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(5, item)
        item = QtWidgets.QTableWidgetItem()
        self.twVideo1.setHorizontalHeaderItem(6, item)
        self.verticalLayout_3.addWidget(self.twVideo1)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.pbAddVideo = QtWidgets.QPushButton(self.tab_player_1)
        self.pbAddVideo.setObjectName("pbAddVideo")
        self.horizontalLayout_3.addWidget(self.pbAddVideo)
        self.pbRemoveVideo = QtWidgets.QPushButton(self.tab_player_1)
        self.pbRemoveVideo.setObjectName("pbRemoveVideo")
        self.horizontalLayout_3.addWidget(self.pbRemoveVideo)
        self.pb_use_media_file_name_as_obsid = QtWidgets.QPushButton(self.tab_player_1)
        self.pb_use_media_file_name_as_obsid.setObjectName("pb_use_media_file_name_as_obsid")
        self.horizontalLayout_3.addWidget(self.pb_use_media_file_name_as_obsid)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_15 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_15.setObjectName("horizontalLayout_15")
        self.cbVisualizeSpectrogram = QtWidgets.QCheckBox(self.tab_player_1)
        self.cbVisualizeSpectrogram.setObjectName("cbVisualizeSpectrogram")
        self.horizontalLayout_15.addWidget(self.cbVisualizeSpectrogram)
        self.cb_visualize_waveform = QtWidgets.QCheckBox(self.tab_player_1)
        self.cb_visualize_waveform.setObjectName("cb_visualize_waveform")
        self.horizontalLayout_15.addWidget(self.cb_visualize_waveform)
        self.cb_media_creation_date_as_offset = QtWidgets.QCheckBox(self.tab_player_1)
        self.cb_media_creation_date_as_offset.setEnabled(False)
        self.cb_media_creation_date_as_offset.setObjectName("cb_media_creation_date_as_offset")
        self.horizontalLayout_15.addWidget(self.cb_media_creation_date_as_offset)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_15.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout_15)
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.label_5 = QtWidgets.QLabel(self.tab_player_1)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_12.addWidget(self.label_5)
        self.sb_media_scan_sampling = QtWidgets.QSpinBox(self.tab_player_1)
        self.sb_media_scan_sampling.setMaximum(1000000)
        self.sb_media_scan_sampling.setObjectName("sb_media_scan_sampling")
        self.horizontalLayout_12.addWidget(self.sb_media_scan_sampling)
        self.label_2 = QtWidgets.QLabel(self.tab_player_1)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_12.addWidget(self.label_2)
        self.sb_image_display_duration = QtWidgets.QSpinBox(self.tab_player_1)
        self.sb_image_display_duration.setMinimum(1)
        self.sb_image_display_duration.setMaximum(86400)
        self.sb_image_display_duration.setObjectName("sb_image_display_duration")
        self.horizontalLayout_12.addWidget(self.sb_image_display_duration)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_12.addItem(spacerItem4)
        self.verticalLayout.addLayout(self.horizontalLayout_12)
        self.cbCloseCurrentBehaviorsBetweenVideo = QtWidgets.QCheckBox(self.tab_player_1)
        self.cbCloseCurrentBehaviorsBetweenVideo.setEnabled(False)
        self.cbCloseCurrentBehaviorsBetweenVideo.setObjectName("cbCloseCurrentBehaviorsBetweenVideo")
        self.verticalLayout.addWidget(self.cbCloseCurrentBehaviorsBetweenVideo)
        self.tabWidget.addTab(self.tab_player_1, "")
        self.tab_data_files = QtWidgets.QWidget()
        self.tab_data_files.setObjectName("tab_data_files")
        self.verticalLayout_17 = QtWidgets.QVBoxLayout(self.tab_data_files)
        self.verticalLayout_17.setObjectName("verticalLayout_17")
        self.splitter_5 = QtWidgets.QSplitter(self.tab_data_files)
        self.splitter_5.setOrientation(QtCore.Qt.Vertical)
        self.splitter_5.setObjectName("splitter_5")
        self.layoutWidget_4 = QtWidgets.QWidget(self.splitter_5)
        self.layoutWidget_4.setObjectName("layoutWidget_4")
        self.verticalLayout_15 = QtWidgets.QVBoxLayout(self.layoutWidget_4)
        self.verticalLayout_15.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_15.setObjectName("verticalLayout_15")
        self.label_7 = QtWidgets.QLabel(self.layoutWidget_4)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_15.addWidget(self.label_7)
        self.tw_data_files = QtWidgets.QTableWidget(self.layoutWidget_4)
        self.tw_data_files.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.tw_data_files.setObjectName("tw_data_files")
        self.tw_data_files.setColumnCount(9)
        self.tw_data_files.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(5, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(6, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(7, item)
        item = QtWidgets.QTableWidgetItem()
        self.tw_data_files.setHorizontalHeaderItem(8, item)
        self.verticalLayout_15.addWidget(self.tw_data_files)
        self.layoutWidget_5 = QtWidgets.QWidget(self.splitter_5)
        self.layoutWidget_5.setObjectName("layoutWidget_5")
        self.verticalLayout_16 = QtWidgets.QVBoxLayout(self.layoutWidget_5)
        self.verticalLayout_16.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_16.setObjectName("verticalLayout_16")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.pb_add_data_file = QtWidgets.QPushButton(self.layoutWidget_5)
        self.pb_add_data_file.setObjectName("pb_add_data_file")
        self.horizontalLayout_5.addWidget(self.pb_add_data_file)
        self.pb_remove_data_file = QtWidgets.QPushButton(self.layoutWidget_5)
        self.pb_remove_data_file.setObjectName("pb_remove_data_file")
        self.horizontalLayout_5.addWidget(self.pb_remove_data_file)
        self.pb_view_data_head = QtWidgets.QPushButton(self.layoutWidget_5)
        self.pb_view_data_head.setObjectName("pb_view_data_head")
        self.horizontalLayout_5.addWidget(self.pb_view_data_head)
        self.pb_plot_data = QtWidgets.QPushButton(self.layoutWidget_5)
        self.pb_plot_data.setObjectName("pb_plot_data")
        self.horizontalLayout_5.addWidget(self.pb_plot_data)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem5)
        self.verticalLayout_16.addLayout(self.horizontalLayout_5)
        spacerItem6 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_16.addItem(spacerItem6)
        self.verticalLayout_17.addWidget(self.splitter_5)
        self.tabWidget.addTab(self.tab_data_files, "")
        self.verticalLayout_7.addWidget(self.tabWidget)
        self.sw_observation_type.addWidget(self.pg_media_files)
        self.pg_live = QtWidgets.QWidget()
        self.pg_live.setObjectName("pg_live")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.pg_live)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.label_4 = QtWidgets.QLabel(self.pg_live)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_8.addWidget(self.label_4)
        self.sbScanSampling = QtWidgets.QSpinBox(self.pg_live)
        self.sbScanSampling.setMaximum(1000000)
        self.sbScanSampling.setObjectName("sbScanSampling")
        self.horizontalLayout_8.addWidget(self.sbScanSampling)
        self.label_6 = QtWidgets.QLabel(self.pg_live)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_8.addWidget(self.label_6)
        spacerItem7 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem7)
        self.verticalLayout_8.addLayout(self.horizontalLayout_8)
        self.cb_start_from_current_time = QtWidgets.QCheckBox(self.pg_live)
        self.cb_start_from_current_time.setObjectName("cb_start_from_current_time")
        self.verticalLayout_8.addWidget(self.cb_start_from_current_time)
        self.rb_day_time = QtWidgets.QRadioButton(self.pg_live)
        self.rb_day_time.setEnabled(False)
        self.rb_day_time.setChecked(True)
        self.rb_day_time.setObjectName("rb_day_time")
        self.verticalLayout_8.addWidget(self.rb_day_time)
        self.rb_epoch_time = QtWidgets.QRadioButton(self.pg_live)
        self.rb_epoch_time.setEnabled(False)
        self.rb_epoch_time.setObjectName("rb_epoch_time")
        self.verticalLayout_8.addWidget(self.rb_epoch_time)
        spacerItem8 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_8.addItem(spacerItem8)
        self.verticalLayout_4.addLayout(self.verticalLayout_8)
        self.sw_observation_type.addWidget(self.pg_live)
        self.pg_images = QtWidgets.QWidget()
        self.pg_images.setObjectName("pg_images")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.pg_images)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.lw_images_directory = QtWidgets.QListWidget(self.pg_images)
        self.lw_images_directory.setObjectName("lw_images_directory")
        self.horizontalLayout_9.addWidget(self.lw_images_directory)
        self.verticalLayout_5.addLayout(self.horizontalLayout_9)
        self.lb_images_info = QtWidgets.QLabel(self.pg_images)
        self.lb_images_info.setObjectName("lb_images_info")
        self.verticalLayout_5.addWidget(self.lb_images_info)
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.pb_add_directory = QtWidgets.QPushButton(self.pg_images)
        self.pb_add_directory.setObjectName("pb_add_directory")
        self.horizontalLayout_14.addWidget(self.pb_add_directory)
        self.pb_remove_directory = QtWidgets.QPushButton(self.pg_images)
        self.pb_remove_directory.setObjectName("pb_remove_directory")
        self.horizontalLayout_14.addWidget(self.pb_remove_directory)
        spacerItem9 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_14.addItem(spacerItem9)
        self.verticalLayout_5.addLayout(self.horizontalLayout_14)
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.pb_use_img_dir_as_obsid = QtWidgets.QPushButton(self.pg_images)
        self.pb_use_img_dir_as_obsid.setObjectName("pb_use_img_dir_as_obsid")
        self.horizontalLayout_13.addWidget(self.pb_use_img_dir_as_obsid)
        spacerItem10 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_13.addItem(spacerItem10)
        self.verticalLayout_5.addLayout(self.horizontalLayout_13)
        self.groupBox = QtWidgets.QGroupBox(self.pg_images)
        self.groupBox.setObjectName("groupBox")
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout(self.groupBox)
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.rb_no_time = QtWidgets.QRadioButton(self.groupBox)
        self.rb_no_time.setObjectName("rb_no_time")
        self.horizontalLayout_10.addWidget(self.rb_no_time)
        self.rb_use_exif = QtWidgets.QRadioButton(self.groupBox)
        self.rb_use_exif.setObjectName("rb_use_exif")
        self.horizontalLayout_10.addWidget(self.rb_use_exif)
        self.rb_time_lapse = QtWidgets.QRadioButton(self.groupBox)
        self.rb_time_lapse.setObjectName("rb_time_lapse")
        self.horizontalLayout_10.addWidget(self.rb_time_lapse)
        self.sb_time_lapse = QtWidgets.QDoubleSpinBox(self.groupBox)
        self.sb_time_lapse.setDecimals(3)
        self.sb_time_lapse.setMaximum(86400.0)
        self.sb_time_lapse.setObjectName("sb_time_lapse")
        self.horizontalLayout_10.addWidget(self.sb_time_lapse)
        spacerItem11 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_10.addItem(spacerItem11)
        self.horizontalLayout_11.addLayout(self.horizontalLayout_10)
        self.verticalLayout_5.addWidget(self.groupBox)
        self.sw_observation_type.addWidget(self.pg_images)
        self.verticalLayout_6.addWidget(self.sw_observation_type)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem12 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem12)
        self.pbCancel = QtWidgets.QPushButton(Form)
        self.pbCancel.setObjectName("pbCancel")
        self.horizontalLayout.addWidget(self.pbCancel)
        self.pbSave = QtWidgets.QPushButton(Form)
        self.pbSave.setObjectName("pbSave")
        self.horizontalLayout.addWidget(self.pbSave)
        self.pbLaunch = QtWidgets.QPushButton(Form)
        self.pbLaunch.setObjectName("pbLaunch")
        self.horizontalLayout.addWidget(self.pbLaunch)
        self.verticalLayout_6.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        self.sw_observation_type.setCurrentIndex(2)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "New observation"))
        self.label.setText(_translate("Form", "Observation id"))
        self.lb_star.setText(_translate("Form", "*"))
        self.label_8.setText(_translate("Form", "Date and time"))
        self.dteDate.setDisplayFormat(_translate("Form", "yyyy-MM-dd hh:mm:ss.zzz"))
        self.label_9.setText(_translate("Form", "Description"))
        self.cb_time_offset.setText(_translate("Form", "Time offset"))
        self.lbTimeOffset.setText(_translate("Form", "Time value"))
        self.cb_observation_time_interval.setText(_translate("Form", "Limit observation to a time interval"))
        self.label_3.setText(_translate("Form", "Independent variables"))
        item = self.twIndepVariables.horizontalHeaderItem(0)
        item.setText(_translate("Form", "Variable"))
        item = self.twIndepVariables.horizontalHeaderItem(1)
        item.setText(_translate("Form", "Type"))
        item = self.twIndepVariables.horizontalHeaderItem(2)
        item.setText(_translate("Form", "Value"))
        self.gb_observation_type.setTitle(_translate("Form", "Observation type"))
        self.rb_media_files.setText(_translate("Form", "Observation from media file(s)"))
        self.rb_live.setText(_translate("Form", "Live observation"))
        self.rb_images.setText(_translate("Form", "Observation from pictures"))
        item = self.twVideo1.horizontalHeaderItem(0)
        item.setText(_translate("Form", "Player"))
        item = self.twVideo1.horizontalHeaderItem(1)
        item.setText(_translate("Form", "Offset (seconds)"))
        item = self.twVideo1.horizontalHeaderItem(2)
        item.setText(_translate("Form", "Path"))
        item = self.twVideo1.horizontalHeaderItem(3)
        item.setText(_translate("Form", "Duration"))
        item = self.twVideo1.horizontalHeaderItem(4)
        item.setText(_translate("Form", "FPS"))
        item = self.twVideo1.horizontalHeaderItem(5)
        item.setText(_translate("Form", "Video"))
        item = self.twVideo1.horizontalHeaderItem(6)
        item.setText(_translate("Form", "Audio"))
        self.pbAddVideo.setText(_translate("Form", "Add media"))
        self.pbRemoveVideo.setText(_translate("Form", "Remove selected media"))
        self.pb_use_media_file_name_as_obsid.setText(_translate("Form", "Use media file name as observation id"))
        self.cbVisualizeSpectrogram.setText(_translate("Form", "Visualize the sound spectrogram for the player #1"))
        self.cb_visualize_waveform.setText(_translate("Form", "Visualize the waveform for the player #1"))
        self.cb_media_creation_date_as_offset.setText(_translate("Form", "Use the media creation date/time metadata as offset"))
        self.label_5.setText(_translate("Form", "Scan sampling every (s)"))
        self.label_2.setText(_translate("Form", "Image display duration (s)"))
        self.cbCloseCurrentBehaviorsBetweenVideo.setText(_translate("Form", "Stop ongoing state events between successive media files"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_player_1), _translate("Form", "Media files"))
        self.label_7.setText(_translate("Form", "Data files to plot"))
        item = self.tw_data_files.horizontalHeaderItem(0)
        item.setText(_translate("Form", "Path"))
        item = self.tw_data_files.horizontalHeaderItem(1)
        item.setText(_translate("Form", "Columns to plot"))
        item = self.tw_data_files.horizontalHeaderItem(2)
        item.setText(_translate("Form", "Plot title"))
        item = self.tw_data_files.horizontalHeaderItem(3)
        item.setText(_translate("Form", "Variable name"))
        item = self.tw_data_files.horizontalHeaderItem(4)
        item.setText(_translate("Form", "Converters"))
        item = self.tw_data_files.horizontalHeaderItem(5)
        item.setText(_translate("Form", "Time interval (s)"))
        item = self.tw_data_files.horizontalHeaderItem(6)
        item.setText(_translate("Form", "Start position (s)"))
        item = self.tw_data_files.horizontalHeaderItem(7)
        item.setText(_translate("Form", "Substract first value"))
        item = self.tw_data_files.horizontalHeaderItem(8)
        item.setText(_translate("Form", "Color"))
        self.pb_add_data_file.setText(_translate("Form", "Add data file"))
        self.pb_remove_data_file.setText(_translate("Form", "Remove selected data file"))
        self.pb_view_data_head.setText(_translate("Form", "View data from file"))
        self.pb_plot_data.setText(_translate("Form", "Show plot"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_data_files), _translate("Form", "Data files"))
        self.label_4.setText(_translate("Form", "Scan sampling every"))
        self.label_6.setText(_translate("Form", "seconds"))
        self.cb_start_from_current_time.setText(_translate("Form", "Start from current time"))
        self.rb_day_time.setText(_translate("Form", "Day time"))
        self.rb_epoch_time.setText(_translate("Form", "Epoch time (seconds since 1970-01-01)"))
        self.lb_images_info.setText(_translate("Form", "Image info:"))
        self.pb_add_directory.setText(_translate("Form", "Add directory"))
        self.pb_remove_directory.setText(_translate("Form", "Remove directory"))
        self.pb_use_img_dir_as_obsid.setText(_translate("Form", "Use the pictures directory as observation id"))
        self.groupBox.setTitle(_translate("Form", "Time"))
        self.rb_no_time.setText(_translate("Form", "No time"))
        self.rb_use_exif.setText(_translate("Form", "Use the EXIF DateTimeOriginal tag"))
        self.rb_time_lapse.setText(_translate("Form", "Time lapse (s)"))
        self.pbCancel.setText(_translate("Form", "Cancel"))
        self.pbSave.setText(_translate("Form", "Save"))
        self.pbLaunch.setText(_translate("Form", "Start"))



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

from PyQt5.QtCore import Qt, pyqtSignal, QEvent, QRect
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QWidget, QPushButton, QHBoxLayout, QGridLayout, QComboBox, QMessageBox

from . import config as cfg
from . import utilities as util


class Button(QWidget):
    def __init__(self, parent=None):
        super(Button, self).__init__(parent)
        self.pushButton = QPushButton()
        self.pushButton.setFocusPolicy(Qt.NoFocus)
        layout = QHBoxLayout()
        layout.addWidget(self.pushButton)
        self.setLayout(layout)


class CodingPad(QWidget):
    clickSignal = pyqtSignal(str)
    sendEventSignal = pyqtSignal(QEvent)
    close_signal = pyqtSignal(QRect, dict)

    def __init__(self, pj: dict, filtered_behaviors, parent=None):
        super().__init__(parent)
        self.pj: dict = pj
        self.filtered_behaviors = filtered_behaviors

        self.behavioral_category_colors_list: list = []
        self.behavior_colors_list: list = []

        self.behavioral_category_colors: dict = {}
        self.behavior_colors: dict = {}

        self.preferences: dict = {"button font size": 20, "button color": cfg.BEHAVIOR_CATEGORY}

        self.button_css: str = "min-width: 50px; min-height:50px; font-weight: bold; max-height:5000px; max-width: 5000px;"

        self.setWindowTitle("Coding pad")

        self.grid = QGridLayout(self)

        self.installEventFilter(self)

    def config(self):
        """
        Configure the coding pad
        """
        if self.cb_config.currentIndex() == 1:  # increase text size
            self.preferences["button font size"] += 4
        if self.cb_config.currentIndex() == 2:  # decrease text size
            self.preferences["button font size"] -= 4
        if self.cb_config.currentIndex() == 3:
            self.preferences["button color"] = cfg.BEHAVIOR_CATEGORY
        if self.cb_config.currentIndex() == 4:
            self.preferences["button color"] = "behavior"
        if self.cb_config.currentIndex() == 5:
            self.preferences["button color"] = "no color"

        self.cb_config.setCurrentIndex(0)
        self.button_configuration()

    def compose(self):
        """
        Add buttons to coding pad
        """
        for i in reversed(range(self.grid.count())):
            if self.grid.itemAt(i).widget() is not None:
                self.grid.itemAt(i).widget().setParent(None)

        # combobox for coding pad configuration
        vlayout = QHBoxLayout()
        self.cb_config = QComboBox()
        self.cb_config.setFocusPolicy(Qt.NoFocus)
        self.cb_config.addItems(
            [
                "Choose an option to configure",
                "Increase button text size",
                "Decrease button text size",
                "Color button by behavioral category",
                "Color button by behavior",
                "No color",
            ]
        )
        self.cb_config.currentIndexChanged.connect(self.config)
        vlayout.addWidget(self.cb_config)
        self.grid.addLayout(vlayout, 0, 1, 1, 1)

        self.all_behaviors = [self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in util.sorted_keys(self.pj[cfg.ETHOGRAM])]

        # behavioral category colors
        self.unique_behavioral_categories = sorted(
            set([self.pj[cfg.ETHOGRAM][x].get(cfg.BEHAVIOR_CATEGORY, "") for x in self.pj[cfg.ETHOGRAM]])
        )
        for idx, category in enumerate(self.unique_behavioral_categories):
            self.behavioral_category_colors[category] = self.behavioral_category_colors_list[
                idx % len(self.behavioral_category_colors_list)
            ]

        # sorted list of unique behavior categories
        behaviorsList = [
            [self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CATEGORY], self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE]]
            for x in util.sorted_keys(self.pj[cfg.ETHOGRAM])
            if cfg.BEHAVIOR_CATEGORY in self.pj[cfg.ETHOGRAM][x] and self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] in self.filtered_behaviors
        ]

        # square grid dimension
        dim = int(len(behaviorsList) ** 0.5 + 0.999)

        c = 0
        for i in range(1, dim + 1):
            for j in range(1, dim + 1):
                if c >= len(behaviorsList):
                    break
                self.addWidget(behaviorsList[c][1], i, j)
                c += 1

        self.button_configuration()

    def addWidget(self, behavior_code: str, i: int, j: int) -> None:
        self.grid.addWidget(Button(), i, j)
        index = self.grid.count() - 1
        widget = self.grid.itemAt(index).widget()

        if widget is not None:
            widget.pushButton.setText(behavior_code)
            widget.pushButton.clicked.connect(lambda: self.click(behavior_code))

    def button_configuration(self):
        """
        configure the font and color of buttons
        """

        state_behaviors_list = util.state_behavior_codes(self.pj[cfg.ETHOGRAM])

        for index in range(self.grid.count()):
            if self.grid.itemAt(index).widget() is None:
                continue
            behavior_code = self.grid.itemAt(index).widget().pushButton.text()

            if self.preferences["button color"] == cfg.BEHAVIOR_CATEGORY:
                behav_cat = [
                    self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CATEGORY]
                    for x in self.pj[cfg.ETHOGRAM]
                    if self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] == behavior_code
                ][0]
                if cfg.BEHAVIORAL_CATEGORIES_CONF in self.pj:
                    behav_cat_color = util.behav_category_user_color(self.pj[cfg.BEHAVIORAL_CATEGORIES_CONF], behav_cat)
                else:
                    behav_cat_color = None
                if behav_cat_color is None:
                    color = self.behavioral_category_colors[behav_cat]
                else:
                    color = behav_cat_color

            if self.preferences["button color"] == "behavior":
                # behavioral categories are not defined
                behavior_position = int(
                    [x for x in util.sorted_keys(self.pj[cfg.ETHOGRAM]) if self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] == behavior_code][0]
                )

                # behavior button color
                behav_color = util.behavior_user_color(self.pj[cfg.ETHOGRAM], behavior_code)

                if behav_color is not None:
                    color = behav_color
                else:
                    color = self.behavior_colors_list[behavior_position % len(self.behavior_colors_list)].replace("tab:", "")

            if self.preferences["button color"] == "no color":
                color = ""

            # set checkable if state behavior
            self.grid.itemAt(index).widget().pushButton.setCheckable(behavior_code in state_behaviors_list)
            self.grid.itemAt(index).widget().pushButton.setStyleSheet(self.button_css + (f"background-color: {color};" if color else ""))
            font = QFont("Arial", self.preferences["button font size"])
            self.grid.itemAt(index).widget().pushButton.setFont(font)

    def resizeEvent(self, event):
        """
        Resize event
        button are redesigned with new font size
        """
        self.button_configuration()

    def click(self, behavior_code: str) -> None:
        """
        Button clicked
        """
        self.clickSignal.emit(behavior_code)

    def eventFilter(self, receiver, event) -> bool:
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.sendEventSignal.emit(event)
            return True
        else:
            return False

    def closeEvent(self, event) -> None:
        """
        send event for widget geometry memory
        """
        self.close_signal.emit(self.geometry(), self.preferences)


def show_coding_pad(self):
    """
    show coding pad window
    """
    if self.playerType in cfg.VIEWERS:
        QMessageBox.warning(self, cfg.programName, "The coding pad is not available in <b>VIEW</b> mode")
        return

    if hasattr(self, "codingpad"):
        self.codingpad.filtered_behaviors = [self.twEthogram.item(i, 1).text() for i in range(self.twEthogram.rowCount())]
        if not self.codingpad.filtered_behaviors:
            QMessageBox.warning(self, cfg.programName, "No behaviors to show!")
            return
        self.codingpad.show()

        # update colors
        self.codingpad.behavior_colors_list = self.plot_colors
        self.codingpad.behavioral_category_colors_list = self.behav_category_colors

    else:  # coding pad does not exist
        filtered_behaviors = [self.twEthogram.item(i, 1).text() for i in range(self.twEthogram.rowCount())]
        if not filtered_behaviors:
            QMessageBox.warning(self, cfg.programName, "No behaviors to show!")
            return
        self.codingpad = CodingPad(self.pj, filtered_behaviors)

        self.codingpad.behavior_colors_list = self.plot_colors
        self.codingpad.behavioral_category_colors_list = self.behav_category_colors

        self.codingpad.compose()

        self.codingpad.setWindowFlags(Qt.WindowStaysOnTopHint)
        self.codingpad.sendEventSignal.connect(self.signal_from_widget)

        print(f"{self.signal_from_widget=}")

        self.codingpad.clickSignal.connect(self.click_signal_from_coding_pad)
        self.codingpad.close_signal.connect(self.close_signal_from_coding_pad)
        self.codingpad.show()

        if self.config_param.get(cfg.CODING_PAD_GEOMETRY, None):
            self.codingpad.setGeometry(
                self.config_param[cfg.CODING_PAD_GEOMETRY].x(),
                self.config_param[cfg.CODING_PAD_GEOMETRY].y(),
                self.config_param[cfg.CODING_PAD_GEOMETRY].width(),
                self.config_param[cfg.CODING_PAD_GEOMETRY].height(),
            )
        else:
            self.codingpad.setGeometry(100, 100, 660, 500)

        if self.config_param.get(cfg.CODING_PAD_CONFIG, {}):
            self.codingpad.preferences = self.config_param[cfg.CODING_PAD_CONFIG]
            self.codingpad.button_configuration()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import math
import statistics
from decimal import Decimal as dec
from typing import Tuple
import tablib
import logging

from . import config as cfg
from . import db_functions
from . import portion as I
from . import project_functions
from . import observation_operations


def default_value(ethogram: dict, behav: str, param):
    """
    return value for duration in case of point event
    """
    default_value_ = 0.0
    behav_type = project_functions.event_type(behav, ethogram)

    if behav_type == "POINT EVENT" and param in (
        "duration",
        "duration mean",
        "duration stdev",
        "proportion of time",
    ):
        default_value_ = cfg.NA

    if behav_type == "STATE EVENT" and param in (
        "duration mean",
        "duration stdev",
    ):
        default_value_ = cfg.NA

    return default_value_


def init_behav_modif(ethogram: dict, selected_subjects: list, distinct_behav_modif: list, parameters: dict) -> dict:
    """
    initialize dictionary with subject, behaviors and modifiers
    """
    behaviors: dict = {}
    for subj in selected_subjects:
        behaviors[subj] = {}
        for behav_modif in distinct_behav_modif:
            behav, modif = behav_modif
            behav_modif_str = "|".join(behav_modif) if modif else behav

            if behav_modif_str not in behaviors[subj]:
                behaviors[subj][behav_modif_str] = {}

            for param in parameters:
                behaviors[subj][behav_modif_str][param[0]] = default_value(ethogram, behav, param[0])

    return behaviors


def init_behav_modif_bin(ethogram: dict, selected_subjects: list, distinct_behav_modif: list, parameters: dict) -> dict:
    """
    initialize dictionary with subject, behaviors and modifiers
    """
    behaviors: dict = {}
    for subj in selected_subjects:
        behaviors[subj] = {}
        for behav_modif in distinct_behav_modif:
            if behav_modif not in behaviors[subj]:
                behaviors[subj][behav_modif] = {}

            for param in parameters:
                behaviors[subj][behav_modif][param[0]] = default_value(ethogram, behav_modif[0], param[0])

    return behaviors


class StdevFunc:
    """
    class to enable std dev function in SQL
    """

    def __init__(self):
        self.M = 0.0
        self.S = 0.0
        self.k = 1

    def step(self, value):
        if value is None:
            return
        tM = self.M
        self.M += (value - tM) / self.k
        self.S += (value - tM) * (value - self.M)
        self.k += 1

    def finalize(self):
        if self.k < 3:
            return None
        return math.sqrt(self.S / (self.k - 2))


def synthetic_time_budget_bin(pj: dict, selected_observations: list, parameters_obs: dict):
    """
    create a synthetic time budget divised in time bin

    Args:
        pj (dict): project dictionary
        selected_observations (list): list of observations to include in time budget
        parameters_obs (dict):

    Returns:
        bool: True if everything OK
        str: message
        tablib.Dataset: dataset containing synthetic time budget data
    """

    def interval_len(interval):
        if interval.empty:
            return dec(0)
        else:
            return sum([x.upper - x.lower for x in interval])

    def interval_number(interval):
        if interval.empty:
            return dec(0)
        else:
            return len(interval)

    def interval_mean(interval):
        if interval.empty:
            return dec(0)
        else:
            return sum([x.upper - x.lower for x in interval]) / len(interval)

    def interval_std_dev(interval) -> str:
        if interval.empty:
            return "NA"
        else:
            try:
                return f"{statistics.stdev([x.upper - x.lower for x in interval]):.3f}"
            except Exception:
                return cfg.NA

    selected_subjects = parameters_obs[cfg.SELECTED_SUBJECTS]
    selected_behaviors = parameters_obs[cfg.SELECTED_BEHAVIORS]
    include_modifiers = parameters_obs[cfg.INCLUDE_MODIFIERS]
    time_interval = parameters_obs["time"]
    start_time = parameters_obs[cfg.START_TIME]
    end_time = parameters_obs[cfg.END_TIME]
    time_bin_size = dec(parameters_obs[cfg.TIME_BIN_SIZE])

    parameters = [
        ["duration", "Total duration"],
        ["number", "Number of occurrences"],
        ["duration mean", "Duration mean"],
        ["duration stdev", "Duration std dev"],
        ["proportion of time", "Proportion of time"],
    ]

    data_report = tablib.Dataset()
    data_report.title = "Synthetic time budget with time bin"

    distinct_behav_modif = []
    for obs_id in selected_observations:
        for event in pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]:
            if include_modifiers:
                if (
                    event[cfg.EVENT_BEHAVIOR_FIELD_IDX],
                    event[cfg.EVENT_MODIFIER_FIELD_IDX],
                ) not in distinct_behav_modif:
                    distinct_behav_modif.append((event[cfg.EVENT_BEHAVIOR_FIELD_IDX], event[cfg.EVENT_MODIFIER_FIELD_IDX]))
            else:
                if (event[cfg.EVENT_BEHAVIOR_FIELD_IDX], "") not in distinct_behav_modif:
                    distinct_behav_modif.append((event[cfg.EVENT_BEHAVIOR_FIELD_IDX], ""))

    distinct_behav_modif.sort()

    # add selected behaviors that are not observed
    for behav in selected_behaviors:
        if [x for x in distinct_behav_modif if x[0] == behav] == []:
            distinct_behav_modif.append((behav, ""))

    param_header = ["Observations id", "Total length (s)", "Time interval (s)"]
    subj_header, behav_header, modif_header = (
        [""] * len(param_header),
        [""] * len(param_header),
        [""] * len(param_header),
    )
    subj_header[1] = "Subjects:"
    behav_header[1] = "Behaviors:"
    modif_header[1] = "Modifiers:"

    for subj in selected_subjects:
        for behavior_modifiers in distinct_behav_modif:
            behavior, modifiers = behavior_modifiers
            behavior_modifiers_str = "|".join(behavior_modifiers) if modifiers else behavior
            for param in parameters:
                subj_header.append(subj)
                behav_header.append(behavior)
                modif_header.append(modifiers)
                param_header.append(param[1])

    data_report.append(subj_header)
    data_report.append(behav_header)
    if include_modifiers:
        data_report.append(modif_header)
    data_report.append(param_header)

    state_events_list = [
        pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in pj[cfg.ETHOGRAM] if cfg.STATE in pj[cfg.ETHOGRAM][x][cfg.TYPE].upper()
    ]
    # select time interval
    for obs_id in selected_observations:
        behaviors = init_behav_modif_bin(pj[cfg.ETHOGRAM], selected_subjects, distinct_behav_modif, parameters)

        obs_length = observation_operations.observation_total_length(pj[cfg.OBSERVATIONS][obs_id])

        if obs_length == -1:
            obs_length = 0
        if time_interval == cfg.TIME_FULL_OBS:
            min_time = dec(0)
            max_time = dec(obs_length)

        if time_interval == cfg.TIME_EVENTS:
            try:
                min_time = dec(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][0][0])
            except Exception:
                min_time = dec(0)
            try:
                max_time = dec(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0])
            except Exception:
                max_time = dec(obs_length)

        if time_interval == cfg.TIME_ARBITRARY_INTERVAL:
            min_time = dec(start_time)
            max_time = dec(end_time)

        events_interval = {}
        mem_events_interval = {}

        for event in pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]:
            if event[cfg.EVENT_SUBJECT_FIELD_IDX] == "":
                current_subject = cfg.NO_FOCAL_SUBJECT
            else:
                current_subject = event[cfg.EVENT_SUBJECT_FIELD_IDX]

            if current_subject not in selected_subjects:
                continue
            if current_subject not in events_interval:
                events_interval[current_subject] = {}
                mem_events_interval[current_subject] = {}

            if include_modifiers:
                modif = event[cfg.EVENT_MODIFIER_FIELD_IDX]
            else:
                modif = ""
            if (event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif) not in distinct_behav_modif:
                continue

            if (event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif) not in events_interval[current_subject]:
                events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)] = I.empty()
                mem_events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)] = []

            if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] in state_events_list:
                mem_events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)].append(event[cfg.EVENT_TIME_FIELD_IDX])
                if len(mem_events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)]) == 2:
                    events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)] |= I.closedopen(
                        mem_events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)][0],
                        mem_events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)][1],
                    )
                    mem_events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)] = []
            else:
                events_interval[current_subject][(event[cfg.EVENT_BEHAVIOR_FIELD_IDX], modif)] |= I.singleton(
                    event[cfg.EVENT_TIME_FIELD_IDX]
                )

        time_bin_start = min_time

        if time_bin_size:
            time_bin_end = time_bin_start + time_bin_size
            if time_bin_end > max_time:
                time_bin_end = max_time
        else:
            time_bin_end = max_time

        while True:
            for subject in events_interval:
                # check behavior to exclude from total time
                time_to_subtract = 0
                if cfg.EXCLUDED_BEHAVIORS in parameters_obs:
                    for behav in events_interval[subject]:
                        if behav[0] in parameters_obs.get(cfg.EXCLUDED_BEHAVIORS, []):
                            interval_intersec = events_interval[subject][behav] & I.closed(time_bin_start, time_bin_end)
                            time_to_subtract += interval_len(interval_intersec)

                for behav in events_interval[subject]:
                    interval_intersec = events_interval[subject][behav] & I.closed(time_bin_start, time_bin_end)

                    nocc = interval_number(interval_intersec)
                    behaviors[subject][behav]["number"] = nocc

                    behav_type = project_functions.event_type(behav[0], pj[cfg.ETHOGRAM])
                    if behav_type == "STATE EVENT":
                        dur = interval_len(interval_intersec)
                        behaviors[subject][behav]["duration"] = f"{dur:.3f}"
                        behaviors[subject][behav]["duration mean"] = f"{interval_mean(interval_intersec):.3f}"
                        behaviors[subject][behav]["duration stdev"] = interval_std_dev(interval_intersec)

                        if behav[0] in parameters_obs.get(cfg.EXCLUDED_BEHAVIORS, []):
                            proportion = dur / (time_bin_end - time_bin_start)
                        else:
                            proportion = dur / ((time_bin_end - time_bin_start) - time_to_subtract)
                        behaviors[subject][behav]["proportion of time"] = f"{proportion:.3f}"

                    if behav_type == "POINT EVENT":
                        behaviors[subject][behav]["duration"] = cfg.NA
                        behaviors[subject][behav]["duration mean"] = cfg.NA
                        behaviors[subject][behav]["duration stdev"] = cfg.NA
                        behaviors[subject][behav]["proportion of time"] = cfg.NA

            columns = [obs_id, f"{max_time - min_time:.3f}", f"{time_bin_start:.3f}-{time_bin_end:.3f}"]
            for subject in selected_subjects:
                for behavior_modifiers in distinct_behav_modif:
                    behavior, modifiers = behavior_modifiers
                    behavior_modifiers_str = behavior_modifiers

                    for param in parameters:
                        columns.append(behaviors[subject][behavior_modifiers_str][param[0]])

            data_report.append(columns)

            time_bin_start = time_bin_end
            time_bin_end = time_bin_start + time_bin_size
            if time_bin_end > max_time:
                time_bin_end = max_time

            if time_bin_start == time_bin_end:
                break

    return True, data_report


def synthetic_time_budget(pj: dict, selected_observations: list, parameters_obs: dict):
    """
    create a synthetic time budget

    Args:
        pj (dict): project dictionary
        selected_observations (list): list of observations to include in time budget
        parameters_obs (dict):

    Returns:
        bool: True if everything OK
        str: message
        tablib.Dataset: dataset containing synthetic time budget data
    """

    selected_subjects = parameters_obs[cfg.SELECTED_SUBJECTS]
    selected_behaviors = parameters_obs[cfg.SELECTED_BEHAVIORS]
    include_modifiers = parameters_obs[cfg.INCLUDE_MODIFIERS]
    interval = parameters_obs["time"]
    start_time = parameters_obs["start time"]
    end_time = parameters_obs["end time"]

    parameters = [
        ["duration", "Total duration"],
        ["number", "Number of occurrences"],
        ["duration mean", "Duration mean"],
        ["duration stdev", "Duration std dev"],
        ["proportion of time", "Proportion of time"],
    ]

    data_report = tablib.Dataset()
    data_report.title = "Synthetic time budget"

    ok, msg, db_connector = db_functions.load_aggregated_events_in_db(pj, selected_subjects, selected_observations, selected_behaviors)

    if not ok:
        return False, msg, None

    db_connector.create_aggregate("stdev", 1, StdevFunc)
    cursor = db_connector.cursor()

    # modifiers
    if include_modifiers:
        cursor.execute("SELECT distinct behavior, modifiers FROM aggregated_events")
        distinct_behav_modif = [[rows["behavior"], rows["modifiers"]] for rows in cursor.fetchall()]
    else:
        cursor.execute("SELECT distinct behavior FROM aggregated_events")
        distinct_behav_modif = [[rows["behavior"], ""] for rows in cursor.fetchall()]

    # add selected behaviors that are not observed
    for behav in selected_behaviors:
        if [x for x in distinct_behav_modif if x[0] == behav] == []:
            distinct_behav_modif.append([behav, ""])

    """behaviors = init_behav_modif(pj[cfg.ETHOGRAM], selected_subjects, distinct_behav_modif, parameters)"""

    param_header = ["Observations id", "Total length (s)"]
    subj_header, behav_header, modif_header = (
        [""] * len(param_header),
        [""] * len(param_header),
        [""] * len(param_header),
    )
    subj_header[1] = "Subjects:"
    behav_header[1] = "Behaviors:"
    modif_header[1] = "Modifiers:"

    for subj in selected_subjects:
        for behavior_modifiers in distinct_behav_modif:
            behavior, modifiers = behavior_modifiers
            behavior_modifiers_str = "|".join(behavior_modifiers) if modifiers else behavior
            for param in parameters:
                subj_header.append(subj)
                behav_header.append(behavior)
                modif_header.append(modifiers)
                param_header.append(param[1])

    data_report.append(subj_header)
    data_report.append(behav_header)
    if include_modifiers:
        data_report.append(modif_header)
    data_report.append(param_header)

    # select time interval
    for obs_id in selected_observations:
        behaviors = init_behav_modif(pj[cfg.ETHOGRAM], selected_subjects, distinct_behav_modif, parameters)

        ok, msg, db_connector = db_functions.load_aggregated_events_in_db(pj, selected_subjects, [obs_id], selected_behaviors)

        if not ok:
            return False, msg, None

        db_connector.create_aggregate("stdev", 1, StdevFunc)
        cursor = db_connector.cursor()
        # if modifiers not to be included set modifiers to ""
        if not include_modifiers:
            cursor.execute("UPDATE aggregated_events SET modifiers = ''")

        # time
        obs_length = observation_operations.observation_total_length(pj[cfg.OBSERVATIONS][obs_id])

        if obs_length == dec(-1):  # media length not available
            interval = cfg.TIME_EVENTS

        if obs_length == dec(-2):  # images obs without time
            interval = cfg.TIME_EVENTS

        if interval == cfg.TIME_FULL_OBS:
            min_time = float(0)
            max_time = float(obs_length)

        if interval == cfg.TIME_EVENTS:
            try:
                min_time = float(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][0][0])
            except Exception:
                min_time = float(0)
            try:
                max_time = float(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0])
            except Exception:
                max_time = float(obs_length)

        if interval == cfg.TIME_ARBITRARY_INTERVAL:
            min_time = float(start_time)
            max_time = float(end_time)

        if obs_length != dec(-2):  # # obs not an images obs without time
            # adapt start and stop to the selected time interval
            cursor.execute(
                "UPDATE aggregated_events SET start = ? WHERE observation = ? AND start < ? AND stop BETWEEN ? AND ?",
                (
                    min_time,
                    obs_id,
                    min_time,
                    min_time,
                    max_time,
                ),
            )
            cursor.execute(
                "UPDATE aggregated_events SET stop = ? WHERE observation = ? AND stop > ? AND start BETWEEN ? AND ?",
                (
                    max_time,
                    obs_id,
                    max_time,
                    min_time,
                    max_time,
                ),
            )

            cursor.execute(
                "UPDATE aggregated_events SET start = ?, stop = ? WHERE observation = ? AND start < ? AND stop > ?",
                (
                    min_time,
                    max_time,
                    obs_id,
                    min_time,
                    max_time,
                ),
            )

            cursor.execute(
                "DELETE FROM aggregated_events WHERE observation = ? AND (start < ? AND stop < ?) OR (start > ? AND stop > ?)",
                (
                    obs_id,
                    min_time,
                    min_time,
                    max_time,
                    max_time,
                ),
            )

        for subject in selected_subjects:
            # check if behaviors are to exclude from total time
            time_to_subtract = 0
            if obs_length != dec(-2):  # obs not an images obs without time
                if cfg.EXCLUDED_BEHAVIORS in parameters_obs:
                    for excluded_behav in parameters_obs[cfg.EXCLUDED_BEHAVIORS]:
                        cursor.execute(
                            ("SELECT SUM(stop-start) " "FROM aggregated_events " "WHERE observation = ? AND subject = ? AND behavior = ? "),
                            (
                                obs_id,
                                subject,
                                excluded_behav,
                            ),
                        )
                        for row in cursor.fetchall():
                            if row[0] is not None:
                                time_to_subtract += row[0]

            for behavior_modifiers in distinct_behav_modif:
                behavior, modifiers = behavior_modifiers
                behavior_modifiers_str = "|".join(behavior_modifiers) if modifiers else behavior

                cursor.execute(
                    (
                        "SELECT SUM(stop - start) AS duration, "
                        "COUNT(*) AS n_occurences, "
                        "AVG(stop - start) AS mean, "
                        "stdev(stop - start) AS ST_DEV, type "
                        "FROM aggregated_events "
                        "WHERE observation = ? AND subject = ? AND behavior = ? AND modifiers = ? "
                    ),
                    (
                        obs_id,
                        subject,
                        behavior,
                        modifiers,
                    ),
                )

                for row in cursor.fetchall():
                    behaviors[subject][behavior_modifiers_str]["number"] = 0 if row["n_occurences"] is None else row["n_occurences"]

                    if obs_length == dec(-2):  # images obs without time
                        behaviors[subject][behavior_modifiers_str]["duration"] = cfg.NA
                        behaviors[subject][behavior_modifiers_str]["duration mean"] = cfg.NA
                        behaviors[subject][behavior_modifiers_str]["duration stdev"] = cfg.NA
                        behaviors[subject][behavior_modifiers_str]["proportion of time"] = cfg.NA

                    else:
                        if row["type"] == cfg.POINT:
                            behaviors[subject][behavior_modifiers_str]["duration"] = cfg.NA
                            behaviors[subject][behavior_modifiers_str]["duration mean"] = cfg.NA
                            behaviors[subject][behavior_modifiers_str]["duration stdev"] = cfg.NA
                            behaviors[subject][behavior_modifiers_str]["proportion of time"] = cfg.NA

                        if row["type"] == cfg.STATE:
                            behaviors[subject][behavior_modifiers_str]["duration"] = (
                                cfg.NA if row["duration"] is None else f"{row['duration']:.3f}"
                            )
                            behaviors[subject][behavior_modifiers_str]["duration mean"] = (
                                cfg.NA if row["mean"] is None else f"{row['mean']:.3f}"
                            )
                            behaviors[subject][behavior_modifiers_str]["duration stdev"] = (
                                cfg.NA if row["ST_DEV"] is None else f"{row['ST_DEV']:.3f}"
                            )

                            if behavior not in parameters_obs[cfg.EXCLUDED_BEHAVIORS]:
                                try:
                                    behaviors[subject][behavior_modifiers_str]["proportion of time"] = (
                                        cfg.NA
                                        if row["duration"] is None
                                        else f"{row['duration'] / ((max_time - min_time) - time_to_subtract):.3f}"
                                    )
                                except ZeroDivisionError:
                                    behaviors[subject][behavior_modifiers_str]["proportion of time"] = cfg.NA
                            else:
                                # behavior subtracted
                                behaviors[subject][behavior_modifiers_str]["proportion of time"] = (
                                    cfg.NA if row["duration"] is None else f"{row['duration'] / (max_time - min_time):.3f}"
                                )

        if obs_length == dec(-2):
            columns = [obs_id, cfg.NA]
        else:
            columns = [obs_id, f"{max_time - min_time:0.3f}"]
        for subj in selected_subjects:
            for behavior_modifiers in distinct_behav_modif:
                behavior, modifiers = behavior_modifiers
                behavior_modifiers_str = "|".join(behavior_modifiers) if modifiers else behavior

                for param in parameters:
                    columns.append(behaviors[subj][behavior_modifiers_str][param[0]])

        data_report.append(columns)

    return True, msg, data_report


def time_budget_analysis(
    ethogram: dict, cursor, selected_observations: list, parameters: dict, by_category: bool = False
) -> Tuple[list, dict]:
    """
    extract number of occurrences, total duration, mean ...
    if start_time = 0 and end_time = 0 all events are extracted

    Args:
        ethogram (dict): project ethogram
        cursor: cursor on temporary database
        selected_observations (list): selected observations
        parameters (dict): parameters for analysis
        by_category (bool): True for grouping in behavioral category else False

    Returns:
        list: results
        dict:
    """

    logging.debug("time_budget_analysis function")

    logging.debug(f"{selected_observations=}")
    logging.debug(f"{parameters=}")

    categories: dict = {}
    out: list = []
    for subject in parameters[cfg.SELECTED_SUBJECTS]:
        out_cat: list = []
        categories[subject]: dict = {}

        for behavior in parameters[cfg.SELECTED_BEHAVIORS]:
            if parameters[cfg.INCLUDE_MODIFIERS]:  # with modifiers
                cursor.execute("SELECT DISTINCT modifiers FROM events WHERE subject = ? AND code = ?", (subject, behavior))
                distinct_modifiers = list(cursor.fetchall())

                if not distinct_modifiers:
                    if not parameters[cfg.EXCLUDE_BEHAVIORS]:
                        if cfg.STATE in project_functions.event_type(behavior, ethogram):
                            # check if observation from pictures
                            if parameters["start time"] == dec("0.000") and parameters["end time"] == dec("0.000"):
                                duration = cfg.NA
                            else:
                                duration = 0.000

                            out.append(
                                {
                                    "subject": subject,
                                    "behavior": behavior,
                                    "modifiers": "",
                                    "duration": duration,
                                    "duration_mean": cfg.NA,
                                    "duration_stdev": cfg.NA,
                                    "number": "0",
                                    "inter_duration_mean": cfg.NA,
                                    "inter_duration_stdev": cfg.NA,
                                }
                            )
                        else:  # point
                            out.append(
                                {
                                    "subject": subject,
                                    "behavior": behavior,
                                    "modifiers": "",
                                    "duration": cfg.NA,
                                    "duration_mean": cfg.NA,
                                    "duration_stdev": cfg.NA,
                                    "number": "0",
                                    "inter_duration_mean": cfg.NA,
                                    "inter_duration_stdev": cfg.NA,
                                }
                            )
                    continue

                if cfg.POINT in project_functions.event_type(behavior, ethogram):
                    for modifier in distinct_modifiers:
                        cursor.execute(
                            (
                                "SELECT occurence, observation FROM events "
                                "WHERE subject = ? "
                                "AND code = ? "
                                "AND modifiers = ? "
                                "ORDER BY observation, occurence"
                            ),
                            (subject, behavior, modifier[0]),
                        )

                        rows = cursor.fetchall()

                        if len(selected_observations) == 1:
                            new_rows: list = []
                            for occurence, observation in rows:
                                if occurence is None:
                                    new_rows.append([float("NaN"), observation])
                                else:
                                    new_rows.append([occurence, observation])
                                """
                                if occurence is not None:
                                    new_occurence = max(float(parameters["start time"]), occurence)
                                    new_occurence = min(new_occurence, float(parameters["end time"]))
                                else:
                                    new_occurence = float("NaN")
                                new_rows.append([new_occurence, observation])
                                """
                            rows = list(new_rows)

                        # include behaviors without events
                        if len(rows) == 0:
                            if not parameters[cfg.EXCLUDE_BEHAVIORS]:
                                out.append(
                                    {
                                        "subject": subject,
                                        "behavior": behavior,
                                        "modifiers": "",
                                        "duration": cfg.NA,
                                        "duration_mean": cfg.NA,
                                        "duration_stdev": cfg.NA,
                                        "number": 0,
                                        "inter_duration_mean": cfg.NA,
                                        "inter_duration_stdev": cfg.NA,
                                    }
                                )
                            continue

                        # inter events duration
                        all_event_interdurations = []
                        for idx, row in enumerate(rows):
                            if idx and row[1] == rows[idx - 1][1]:
                                all_event_interdurations.append(float(row[0]) - float(rows[idx - 1][0]))

                        if [x for x in all_event_interdurations if math.isnan(x)] or len(all_event_interdurations) == 0:
                            inter_duration_mean = cfg.NA
                            inter_duration_stdev = cfg.NA
                        else:
                            inter_duration_mean = round(statistics.mean(all_event_interdurations), 3)
                            if len(all_event_interdurations) > 1:
                                inter_duration_stdev = round(statistics.stdev(all_event_interdurations), 3)
                            else:
                                inter_duration_stdev = cfg.NA

                        out_cat.append(
                            {
                                "subject": subject,
                                "behavior": behavior,
                                "modifiers": modifier[0],
                                "duration": cfg.NA,
                                "duration_mean": cfg.NA,
                                "duration_stdev": cfg.NA,
                                "number": len(rows),
                                "inter_duration_mean": inter_duration_mean,
                                "inter_duration_stdev": inter_duration_stdev,
                            }
                        )

                if cfg.STATE in project_functions.event_type(behavior, ethogram):
                    for modifier in distinct_modifiers:
                        cursor.execute(
                            (
                                "SELECT occurence, observation FROM events "
                                "WHERE subject = ? "
                                "AND code = ? "
                                "AND modifiers = ? "
                                "ORDER BY observation, occurence"
                            ),
                            (subject, behavior, modifier[0]),
                        )

                        rows = list(cursor.fetchall())

                        if len(rows) == 0:
                            if not parameters[cfg.EXCLUDE_BEHAVIORS]:  # include behaviors without events
                                # check if observation from pictures
                                if parameters["start time"] == dec("0.000") and parameters["end time"] == dec("0.000"):
                                    duration = cfg.NA
                                else:
                                    duration: float = 0.000
                                out.append(
                                    {
                                        "subject": subject,
                                        "behavior": behavior,
                                        "modifiers": modifier[0],
                                        "duration": duration,
                                        "duration_mean": cfg.NA,
                                        "duration_stdev": cfg.NA,
                                        "number": 0,
                                        "inter_duration_mean": cfg.NA,
                                        "inter_duration_stdev": cfg.NA,
                                    }
                                )
                            continue

                        if len(rows) % 2:
                            out.append(
                                {
                                    "subject": subject,
                                    "behavior": behavior,
                                    "modifiers": modifier[0],
                                    "duration": cfg.UNPAIRED,
                                    "duration_mean": cfg.UNPAIRED,
                                    "duration_stdev": cfg.UNPAIRED,
                                    "number": cfg.UNPAIRED,
                                    "inter_duration_mean": cfg.UNPAIRED,
                                    "inter_duration_stdev": cfg.UNPAIRED,
                                }
                            )
                        else:
                            all_event_durations: list = []
                            all_event_interdurations: list = []
                            for idx, row in enumerate(rows):
                                # event
                                if idx % 2 == 0:
                                    if row[0] is not None and rows[idx + 1][0] is not None:
                                        new_init, new_end = float(row[0]), float(rows[idx + 1][0])
                                        all_event_durations.append(new_end - new_init)
                                    else:
                                        all_event_durations.append(float("NaN"))

                                # inter event if same observation
                                if idx % 2 and idx != len(rows) - 1 and row[1] == rows[idx + 1][1]:
                                    if row[0] is not None and rows[idx + 1][0] is not None:
                                        # and (
                                        # parameters["start time"] <= row[0] <= parameters["end time"]
                                        # and parameters["start time"] <= rows[idx + 1][0] <= parameters["end time"]
                                        # ):
                                        all_event_interdurations.append(float(rows[idx + 1][0]) - float(row[0]))
                                    else:
                                        all_event_interdurations.append(float("NaN"))

                            # events
                            if [x for x in all_event_durations if math.isnan(x)]:
                                total_duration = cfg.NA
                            else:
                                total_duration = round(sum(all_event_durations), 3)
                            if [x for x in all_event_durations if math.isnan(x)] or len(all_event_durations) == 0:
                                duration_mean = cfg.NA
                                duration_stdev = cfg.NA
                            else:
                                duration_mean = round(statistics.mean(all_event_durations), 3)
                                if len(all_event_durations) > 1:
                                    duration_stdev = round(statistics.stdev(all_event_durations), 3)
                                else:
                                    duration_stdev = cfg.NA
                            # interduration
                            if [x for x in all_event_interdurations if math.isnan(x)] or len(all_event_interdurations) == 0:
                                inter_duration_mean = cfg.NA
                                inter_duration_stdev = cfg.NA
                            else:
                                inter_duration_mean = round(statistics.mean(all_event_interdurations), 3)
                                if len(all_event_interdurations) > 1:
                                    inter_duration_stdev = round(statistics.stdev(all_event_interdurations), 3)
                                else:
                                    inter_duration_stdev = cfg.NA

                            out_cat.append(
                                {
                                    "subject": subject,
                                    "behavior": behavior,
                                    "modifiers": modifier[0],
                                    "duration": total_duration,
                                    "duration_mean": duration_mean,
                                    "duration_stdev": duration_stdev,
                                    "number": len(all_event_durations),
                                    "inter_duration_mean": inter_duration_mean,
                                    "inter_duration_stdev": inter_duration_stdev,
                                }
                            )

            else:  # no modifiers
                if cfg.POINT in project_functions.event_type(behavior, ethogram):
                    cursor.execute(
                        ("SELECT occurence,observation FROM events WHERE subject = ? AND code = ? ORDER BY observation, occurence"),
                        (subject, behavior),
                    )

                    rows = list(cursor.fetchall())

                    if len(selected_observations) == 1:
                        new_rows: list = []
                        for occurence, observation in rows:
                            if occurence is None:
                                new_rows.append([float("NaN"), observation])
                            else:
                                new_rows.append([occurence, observation])
                            """
                            if occurence is not None:
                                new_occurence = max(float(parameters["start time"]), occurence)
                                new_occurence = min(new_occurence, float(parameters["end time"]))
                            else:
                                new_occurence = float("NaN")
                            new_rows.append([new_occurence, observation])
                            """

                        rows = list(new_rows)

                    # include behaviors without events
                    if len(rows) == 0:
                        if not parameters[cfg.EXCLUDE_BEHAVIORS]:
                            out.append(
                                {
                                    "subject": subject,
                                    "behavior": behavior,
                                    "modifiers": "",
                                    "duration": cfg.NA,
                                    "duration_mean": cfg.NA,
                                    "duration_stdev": cfg.NA,
                                    "number": 0,
                                    "inter_duration_mean": cfg.NA,
                                    "inter_duration_stdev": cfg.NA,
                                }
                            )
                        continue

                    # inter events duration
                    all_event_interdurations = []
                    for idx, row in enumerate(rows):
                        if idx and row[1] == rows[idx - 1][1]:
                            all_event_interdurations.append(float(row[0]) - float(rows[idx - 1][0]))

                    if [x for x in all_event_interdurations if math.isnan(x)] or len(all_event_interdurations) == 0:
                        inter_duration_mean = cfg.NA
                        inter_duration_stdev = cfg.NA
                    else:
                        inter_duration_mean = round(statistics.mean(all_event_interdurations), 3)
                        if len(all_event_interdurations) > 1:
                            inter_duration_stdev = round(statistics.stdev(all_event_interdurations), 3)
                        else:
                            inter_duration_stdev = cfg.NA

                    out_cat.append(
                        {
                            "subject": subject,
                            "behavior": behavior,
                            "modifiers": "",
                            "duration": cfg.NA,
                            "duration_mean": cfg.NA,
                            "duration_stdev": cfg.NA,
                            "number": len(rows),
                            "inter_duration_mean": inter_duration_mean,
                            "inter_duration_stdev": inter_duration_stdev,
                        }
                    )

                if cfg.STATE in project_functions.event_type(behavior, ethogram):
                    cursor.execute(
                        ("SELECT occurence, observation FROM events WHERE subject = ? AND code = ? ORDER BY observation, occurence"),
                        (subject, behavior),
                    )

                    rows = list(cursor.fetchall())

                    if len(rows) == 0:
                        if not parameters[cfg.EXCLUDE_BEHAVIORS]:  # include behaviors without events
                            # check if observation from pictures
                            if parameters["start time"] == dec("0.000") and parameters["end time"] == dec("0.000"):
                                duration = cfg.NA
                            else:
                                duration = 0.000
                            out.append(
                                {
                                    "subject": subject,
                                    "behavior": behavior,
                                    "modifiers": "",
                                    "duration": duration,
                                    "duration_mean": cfg.NA,
                                    "duration_stdev": cfg.NA,
                                    "number": 0,
                                    "inter_duration_mean": cfg.NA,
                                    "inter_duration_stdev": cfg.NA,
                                }
                            )
                        continue

                    if len(rows) % 2:  # unpaired events
                        out.append(
                            {
                                "subject": subject,
                                "behavior": behavior,
                                "modifiers": "",
                                "duration": cfg.UNPAIRED,
                                "duration_mean": cfg.UNPAIRED,
                                "duration_stdev": cfg.UNPAIRED,
                                "number": cfg.UNPAIRED,
                                "inter_duration_mean": cfg.UNPAIRED,
                                "inter_duration_stdev": cfg.UNPAIRED,
                            }
                        )
                    else:
                        all_event_durations: list = []
                        all_event_interdurations: list = []
                        for idx, row in enumerate(rows):
                            # event
                            if idx % 2 == 0:
                                if row[0] is not None and rows[idx + 1][0] is not None:
                                    new_init, new_end = float(row[0]), float(rows[idx + 1][0])
                                    all_event_durations.append(new_end - new_init)
                                else:
                                    all_event_durations.append(float("NaN"))

                            # inter event if same observation
                            if idx % 2 and idx != len(rows) - 1 and row[1] == rows[idx + 1][1]:
                                if row[0] is not None and rows[idx + 1][0] is not None:
                                    # and (
                                    # parameters["start time"] <= row[0] <= parameters["end time"]
                                    # and parameters["start time"] <= rows[idx + 1][0] <= parameters["end time"]
                                    # ):
                                    all_event_interdurations.append(float(rows[idx + 1][0]) - float(row[0]))
                                else:
                                    all_event_interdurations.append(float("NaN"))

                        # events
                        if [x for x in all_event_durations if math.isnan(x)]:
                            total_duration = cfg.NA
                        else:
                            total_duration = round(sum(all_event_durations), 3)
                        if [x for x in all_event_durations if math.isnan(x)] or len(all_event_durations) == 0:
                            duration_mean = cfg.NA
                            duration_stdev = cfg.NA
                        else:
                            duration_mean = round(statistics.mean(all_event_durations), 3)
                            if len(all_event_durations) > 1:
                                duration_stdev = round(statistics.stdev(all_event_durations), 3)
                            else:
                                duration_stdev = cfg.NA
                        # interduration
                        if [x for x in all_event_interdurations if math.isnan(x)] or len(all_event_interdurations) == 0:
                            inter_duration_mean = cfg.NA
                            inter_duration_stdev = cfg.NA
                        else:
                            inter_duration_mean = round(statistics.mean(all_event_interdurations), 3)
                            if len(all_event_interdurations) > 1:
                                inter_duration_stdev = round(statistics.stdev(all_event_interdurations), 3)
                            else:
                                inter_duration_stdev = cfg.NA

                        out_cat.append(
                            {
                                "subject": subject,
                                "behavior": behavior,
                                "modifiers": "",
                                "duration": total_duration,
                                "duration_mean": duration_mean,
                                "duration_stdev": duration_stdev,
                                "number": len(all_event_durations),
                                "inter_duration_mean": inter_duration_mean,
                                "inter_duration_stdev": inter_duration_stdev,
                            }
                        )

        out += out_cat

        if by_category:  # and flagCategories:
            for behav in out_cat:
                try:
                    category = [
                        ethogram[x][cfg.BEHAVIOR_CATEGORY]
                        for x in ethogram
                        if cfg.BEHAVIOR_CATEGORY in ethogram[x] and ethogram[x][cfg.BEHAVIOR_CODE] == behav["behavior"]
                    ][0]
                except Exception:
                    category = ""

                if category not in categories[subject]:
                    categories[subject][category] = {"duration": 0, "number": 0}

                if cfg.STATE in project_functions.event_type(behav["behavior"], ethogram):
                    if behav["duration"] not in ("-", cfg.NA) and categories[subject][category]["duration"] not in (
                        "-",
                        cfg.NA,
                    ):
                        categories[subject][category]["duration"] += behav["duration"]
                    else:
                        categories[subject][category]["duration"] = cfg.NA

                categories[subject][category]["number"] += behav["number"]

    out_sorted: list = []
    for subject in parameters[cfg.SELECTED_SUBJECTS]:
        for behavior in parameters[cfg.SELECTED_BEHAVIORS]:
            for row in out:
                if row[cfg.SUBJECT] == subject and row["behavior"] == behavior:
                    out_sorted.append(row)

    # http://stackoverflow.com/questions/673867/python-arbitrary-order-by
    return (out_sorted, categories)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""


import json
import datetime
from pathlib import Path
import pandas as pd

from PyQt5.QtWidgets import (
    QMessageBox,
    QFileDialog,
)

from . import config as cfg
from . import dialog
from . import utilities as util


def load_observations_from_boris_project(self, project_file_path: str):
    """
    import observations from a BORIS project file
    """

    if self.projectFileName and project_file_path == self.projectFileName:
        QMessageBox.critical(
            None,
            cfg.programName,
            "This project is already open",
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    try:
        fromProject = json.loads(open(project_file_path, "r").read())
    except Exception:
        QMessageBox.critical(self, cfg.programName, "This project file seems corrupted")
        return

    # transform time to decimal
    fromProject = util.convert_time_to_decimal(fromProject)  # function in utilities.py

    dbc = dialog.ChooseObservationsToImport("Choose the observations to import:", sorted(list(fromProject[cfg.OBSERVATIONS].keys())))

    if not dbc.exec_():
        return
    selected_observations = dbc.get_selected_observations()
    if selected_observations:
        flagImported = False

        # set of behaviors in current projet ethogram
        behav_set = set([self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE] for idx in self.pj[cfg.ETHOGRAM]])

        # set of subjects in current projet
        subjects_set = set([self.pj[cfg.SUBJECTS][idx][cfg.SUBJECT_NAME] for idx in self.pj[cfg.SUBJECTS]])

        for obs_id in selected_observations:
            # check if behaviors are in current project ethogram
            new_behav_set = set(
                [
                    event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
                    for event in fromProject[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]
                    if event[cfg.EVENT_BEHAVIOR_FIELD_IDX] not in behav_set
                ]
            )
            if new_behav_set:
                diag_result = dialog.MessageDialog(
                    cfg.programName,
                    (f"Some coded behaviors in <b>{obs_id}</b> are " f"not defined in the ethogram:<br><b>{', '.join(new_behav_set)}</b>"),
                    ["Interrupt import", "Skip observation", "Import observation"],
                )
                if diag_result == "Interrupt import":
                    return
                if diag_result == "Skip observation":
                    continue

            # check if subjects are in current project
            new_subject_set = set(
                [
                    event[cfg.EVENT_SUBJECT_FIELD_IDX]
                    for event in fromProject[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]
                    if event[cfg.EVENT_SUBJECT_FIELD_IDX] not in subjects_set
                ]
            )
            if new_subject_set and new_subject_set != {""}:
                diag_result = dialog.MessageDialog(
                    cfg.programName,
                    (f"Some coded subjects in <b>{obs_id}</b> are not defined in the project:<br>" f"<b>{', '.join(new_subject_set)}</b>"),
                    ["Interrupt import", "Skip observation", "Import observation"],
                )

                if diag_result == "Interrupt import":
                    return

                if diag_result == "Skip observation":
                    continue

            if obs_id in self.pj[cfg.OBSERVATIONS].keys():
                diag_result = dialog.MessageDialog(
                    cfg.programName,
                    (f"The observation <b>{obs_id}</b>" "already exists in the current project.<br>"),
                    ["Interrupt import", "Skip observation", "Rename observation"],
                )
                if diag_result == "Interrupt import":
                    return

                if diag_result == "Rename observation":
                    self.pj[cfg.OBSERVATIONS][f"{obs_id} (imported at {util.datetime_iso8601(datetime.datetime.now())})"] = dict(
                        fromProject[cfg.OBSERVATIONS][obs_id]
                    )
                    flagImported = True
            else:
                self.pj[cfg.OBSERVATIONS][obs_id] = dict(fromProject[cfg.OBSERVATIONS][obs_id])
                flagImported = True

        if flagImported:
            QMessageBox.information(self, cfg.programName, "Observations imported successfully")
            self.project_changed()


def load_observations_from_spreadsheet(self, project_file_path: str):
    """
    import observations from a spreadsheet file
    """

    if Path(project_file_path).suffix.upper() == ".XLSX":
        engine = "openpyxl"
    elif Path(project_file_path).suffix.upper() == ".ODS":
        engine = "odf"
    else:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Microsoft-Excel XLSX format or OpenDocument ODS"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    try:
        df = pd.read_excel(project_file_path, sheet_name=0, engine=engine)
    except Exception:
        QMessageBox.warning(
            None,
            cfg.programName,
            ("The type of file was not recognized. Must be Microsoft-Excel XLSX format or OpenDocument ODS"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    expected_labels: list = ["time", "subject", "code", "modifier", "comment"]

    df.columns = df.columns.str.upper()

    for column in expected_labels:
        if column.upper() not in list(df.columns):
            QMessageBox.warning(
                None,
                cfg.programName,
                (
                    f"The {column} column was not found in the file header.<br>"
                    "For information the current file header contains the following labels:<br>"
                    f"{'<br>'.join(['<b>' + util.replace_leading_trailing_chars(x, ' ', '&#9608;') + '</b>' for x in df.columns])}<br>"
                    "<br>"
                    "The first row of the spreadsheet must contain the following labels:<br>"
                    f"{'<br>'.join(['<b>' + x + '</b>' for x in expected_labels])}<br>"
                    "<br>The order is not mandatory."
                ),
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return 1
    event: dict = {}
    events: list = []
    for _, row in df.iterrows():
        for label in expected_labels:
            event[label] = row[label.upper()] if str(row[label.upper()]) != "nan" else ""
        events.append([event["time"], event["subject"], event["code"], event["modifier"], event["comment"]])

    if events:
        self.pj[cfg.OBSERVATIONS][self.observationId]["events"].extend(events)
        self.load_tw_events(self.observationId)

        QMessageBox.information(self, cfg.programName, "Observations imported successfully")
        self.project_changed()


def import_observations(self):
    """
    import observations from project file
    """

    fn = QFileDialog().getOpenFileName(
        None, "Choose a file", "", "BORIS project files (*.boris);;Spreadsheet files (*.ods *.xlsx *);;All files (*)"
    )
    file_name = fn[0] if type(fn) is tuple else fn

    if not file_name:
        return

    if Path(file_name).suffix == ".boris":
        load_observations_from_boris_project(self, file_name)

    if Path(file_name).suffix in (".ods", ".xlsx"):
        if not self.observationId:
            QMessageBox.warning(
                None,
                cfg.programName,
                ("Please open or create a new observation before importing from a spreadsheet file"),
                QMessageBox.Ok,
                QMessageBox.NoButton,
            )
            return

        load_observations_from_spreadsheet(self, file_name)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2020 Olivier Friard

module for testing export_observation.py

pytest -vv test_export_observation.py
"""

import pytest
import sys
import json
import os
from openpyxl import load_workbook
import tablib

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from boris import config
from boris import export_observation


SELECTED_SUBJECTS = config.SELECTED_SUBJECTS
SELECTED_BEHAVIORS = config.SELECTED_BEHAVIORS


@pytest.fixture()
def before():
    os.system("rm -rf output")
    os.system("mkdir output")


class Test_export_events(object):
    @pytest.mark.usefixtures("before")
    def test_export_tabular_tsv(self):

        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #1"
        parameters = {SELECTED_SUBJECTS: ["subject1", "subject2"], SELECTED_BEHAVIORS: ["p", "s"]}
        file_name = "test_export_events_tabular.tsv"
        output_format = "tsv"

        r, msg = export_observation.export_events(
            parameters,
            obs_id,
            pj[config.OBSERVATIONS][obs_id],
            pj[config.ETHOGRAM],
            "output/" + file_name,
            output_format,
        )

        assert (
            open("files/test_export_events_tabular.tsv").read() == open("output/test_export_events_tabular.tsv").read()
        )

    @pytest.mark.usefixtures("before")
    def test_export_tabular_csv(self):

        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #1"
        parameters = {SELECTED_SUBJECTS: ["subject1", "subject2"], SELECTED_BEHAVIORS: ["p", "s"]}
        file_name = "test_export_events_tabular.csv"
        output_format = "csv"

        r, msg = export_observation.export_events(
            parameters,
            obs_id,
            pj[config.OBSERVATIONS][obs_id],
            pj[config.ETHOGRAM],
            "output/" + file_name,
            output_format,
        )

        assert (
            open("files/test_export_events_tabular.csv").read() == open("output/test_export_events_tabular.csv").read()
        )

    @pytest.mark.usefixtures("before")
    def test_export_tabular_html(self):

        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #1"
        parameters = {SELECTED_SUBJECTS: ["subject1", "subject2"], SELECTED_BEHAVIORS: ["p", "s"]}
        file_name = "test_export_events_tabular.html"
        output_format = "html"

        r, msg = export_observation.export_events(
            parameters,
            obs_id,
            pj[config.OBSERVATIONS][obs_id],
            pj[config.ETHOGRAM],
            "output/" + file_name,
            output_format,
        )

        assert (
            open("files/test_export_events_tabular.html").read()
            == open("output/test_export_events_tabular.html").read()
        )

    @pytest.mark.usefixtures("before")
    def test_export_tabular_xlsx(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "observation #1"
        parameters = {SELECTED_SUBJECTS: ["subject1", "subject2"], SELECTED_BEHAVIORS: ["p", "s"]}
        file_name = "test_export_events_tabular.xlsx"
        output_format = "xlsx"

        r, msg = export_observation.export_events(
            parameters,
            obs_id,
            pj[config.OBSERVATIONS][obs_id],
            pj[config.ETHOGRAM],
            "output/" + file_name,
            output_format,
        )

        ref_all_cells = []
        wb = load_workbook(filename=f"files/{file_name}", read_only=True)
        for ws_name in wb.sheetnames:
            ref_all_cells.extend([cell.value for row in wb[ws_name].rows for cell in row])

        test_all_cells = []
        wb = load_workbook(filename=f"output/{file_name}", read_only=True)
        for ws_name in wb.sheetnames:
            worksheet_name = ws_name
            test_all_cells.extend([cell.value for row in wb[ws_name].rows for cell in row])

        assert worksheet_name == obs_id
        assert ref_all_cells == test_all_cells


# TODO: add testing long worksheet title xls

# TODO: add testing for worksheet title containing forbidden characters xls/xlsx


class Test_export_aggregated_events(object):
    @pytest.mark.usefixtures("before")
    def test_full_1(self):
        """
        export aggregated events:
        no time interval restriction: 9 min
        all subjects
        all behaviors
        """
        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #2"
        parameters = {
            "selected subjects": ["subject1", "subject2", "No focal subject"],
            SELECTED_BEHAVIORS: ["p", "s"],
            "time": config.TIME_FULL_OBS,
            config.START_TIME: 0,
            config.END_TIME: 9 * 60,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        print(tablib_dataset_tsv)

        """
        print(tablib_dataset_tsv)
        open("files/test_export_aggregated_events_test_full_1.tsv","w").write(tablib_dataset_tsv)
        """

        ref = open("files/test_export_aggregated_events_test_full_1.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref

    @pytest.mark.usefixtures("before")
    def test_full_2(self):
        """
        export aggregated events:
        no time interval restriction: 9 min
        1 subject / 2
        all behaviors
        """
        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #2"
        parameters = {
            "selected subjects": ["subject1"],
            "selected behaviors": ["p", "s"],
            "time": config.TIME_FULL_OBS,
            config.START_TIME: 0,
            config.END_TIME: 9 * 60,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        """
        print(tablib_dataset_json)
        open("files/test_export_aggregated_events_test_full_2.tsv","w").write(tablib_dataset_tsv)
        """

        ref = open("files/test_export_aggregated_events_test_full_2.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref

    @pytest.mark.usefixtures("before")
    def test_full_3(self):
        """
        export aggregated events:
        no time interval restriction: 9 min
        all subjects
        1 behavior / 2
        """
        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #2"
        parameters = {
            "selected subjects": ["subject1", "subject2", "No focal subject"],
            "selected behaviors": ["s"],
            "time": config.TIME_FULL_OBS,
            config.START_TIME: 0,
            config.END_TIME: 9 * 60,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        """
        print(tablib_dataset_tsv)
        open("files/test_export_aggregated_events_test_full_3.tsv","w").write(tablib_dataset_tsv)
        """

        ref = open("files/test_export_aggregated_events_test_full_3.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref

    @pytest.mark.usefixtures("before")
    def test_full_with_trailing_spaces_in_modifiers(self):
        """
        some modifiers were configured with trailing spaces
        """
        pj = json.loads(open("files/test_with_leading_trailing_spaces_in_modifiers.boris").read())
        obs_id = "test1 live"
        parameters = {
            "selected subjects": ["No focal subject"],
            "selected behaviors": ["p", "r"],
            "time": config.TIME_FULL_OBS,
            config.START_TIME: 0,
            config.END_TIME: 2 * 60,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        """
        print(tablib_dataset_tsv)
        open("files/test_export_aggregated_events_test_full_with_trailing_spaces_in_modifiers.tsv", "w").write(tablib_dataset_tsv)
        """

        ref = open("files/test_export_aggregated_events_test_full_with_trailing_spaces_in_modifiers.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref

    @pytest.mark.usefixtures("before")
    def test_partial_4(self):
        """
        export aggregated events:
        time interval restriction: 0 - 60 s
        all subjects
        all behaviors
        """
        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #2"
        parameters = {
            "selected subjects": ["subject1", "subject2", "No focal subject"],
            "selected behaviors": ["s", "p"],
            "time": config.TIME_ARBITRARY_INTERVAL,
            config.START_TIME: 0,
            config.END_TIME: 1 * 60,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        """
        print(tablib_dataset_tsv)
        open("files/test_export_aggregated_events_test_full_4.tsv","w").write(tablib_dataset_tsv)
        """

        ref = open("files/test_export_aggregated_events_test_full_4.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref

    @pytest.mark.usefixtures("before")
    def test_partial_5(self):
        """
        export aggregated events:
        time interval restriction: 5 - 30 s
        all subjects
        all behaviors
        """
        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #2"
        parameters = {
            "selected subjects": ["subject1", "subject2", "No focal subject"],
            "selected behaviors": ["s", "p"],
            "time": config.TIME_ARBITRARY_INTERVAL,
            config.START_TIME: 5,
            config.END_TIME: 30,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        print(tablib_dataset_tsv)
        open("files/test_export_aggregated_events_test_full_5.tsv", "w").write(tablib_dataset_tsv)

        ref = open("files/test_export_aggregated_events_test_full_5.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref

    @pytest.mark.usefixtures("before")
    def test_partial_6(self):
        """
        export aggregated events: No events
        time interval restriction: 60 - 180 s
        all subjects
        all behaviors
        """
        pj = json.loads(open("files/test.boris").read())
        obs_id = "observation #2"
        parameters = {
            config.SELECTED_SUBJECTS: ["subject1", "subject2", "No focal subject"],
            config.SELECTED_BEHAVIORS: ["s", "p"],
            "time": config.TIME_ARBITRARY_INTERVAL,
            config.START_TIME: 60,
            config.END_TIME: 180,
        }

        tablib_dataset = export_observation.export_aggregated_events(pj, parameters, obs_id)
        tablib_dataset_tsv = tablib_dataset.tsv

        """
        print(tablib_dataset_tsv)
        open("files/test_export_aggregated_events_test_full_6.tsv","w").write(tablib_dataset_tsv)
        """

        ref = open("files/test_export_aggregated_events_test_full_6.tsv").read()
        assert tablib_dataset_tsv.replace("\r", "") == ref


class Test_export_events_jwatcher(object):
    @pytest.mark.usefixtures("before")
    def test_1(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "observation #1"
        parameters = {config.SELECTED_SUBJECTS: ["subject1"], config.SELECTED_BEHAVIORS: ["p", "s"]}
        file_name = "test_jwatcher"
        output_format = ""

        r, msg = export_observation.export_events_jwatcher(
            parameters, obs_id, pj["observations"][obs_id], pj[config.ETHOGRAM], "output/" + file_name, output_format
        )

        ref = [x for x in open("files/test_jwatcher_subject1.dat").readlines() if not x.startswith("#")]
        out = [x for x in open("output/test_jwatcher_subject1.dat").readlines() if not x.startswith("#")]
        assert ref == out

        ref = [x for x in open("files/test_jwatcher_subject1.faf").readlines() if not x.startswith("#")]
        out = [x for x in open("output/test_jwatcher_subject1.faf").readlines() if not x.startswith("#")]
        assert ref == out

        ref = [x for x in open("files/test_jwatcher_subject1.fmf").readlines() if not x.startswith("#")]
        out = [x for x in open("output/test_jwatcher_subject1.fmf").readlines() if not x.startswith("#")]
        assert ref == out


class Test_events_to_behavioral_sequences(object):
    def test_1(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "observation #1"
        subject = "subject1"
        parameters = {
            SELECTED_SUBJECTS: ["subject1"],
            SELECTED_BEHAVIORS: ["p", "s"],
            config.INCLUDE_MODIFIERS: False,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "|"

        out = export_observation.events_to_behavioral_sequences(pj, obs_id, subject, parameters, behav_seq_separator)

        assert open("files/Test_events_to_behavioral_sequences_test_1").read() == out

    def test_2_separator_changed(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "observation #1"
        subject = "subject1"
        parameters = {
            SELECTED_SUBJECTS: ["subject1"],
            SELECTED_BEHAVIORS: ["p", "s"],
            config.INCLUDE_MODIFIERS: False,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "£"

        out = export_observation.events_to_behavioral_sequences(pj, obs_id, subject, parameters, behav_seq_separator)

        assert open("files/Test_events_to_behavioral_sequences_test_2_separator").read() == out

    def test_3_no_behavior_found_for_selected_subject(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "observation #1"
        subject = "subject1"
        parameters = {
            SELECTED_SUBJECTS: ["subject2"],
            SELECTED_BEHAVIORS: ["p"],
            config.INCLUDE_MODIFIERS: False,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "|"

        out = export_observation.events_to_behavioral_sequences(pj, obs_id, subject, parameters, behav_seq_separator)

        # open("1", "w").write(out)
        assert out == ""

    def test_4_behaviors_with_modifiers(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "modifiers"
        subject = ""
        parameters = {
            SELECTED_SUBJECTS: [""],
            SELECTED_BEHAVIORS: ["q", "r"],
            config.INCLUDE_MODIFIERS: True,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "|"

        out = export_observation.events_to_behavioral_sequences(pj, obs_id, subject, parameters, behav_seq_separator)

        assert open("files/Test_events_to_behavioral_sequences_test_4_behaviors_with_modifiers").read() == out

    def test_5_observation_not_paired(self):

        pj = json.loads(open("files/test.boris").read())

        obs_id = "live not paired"
        subject = ""
        parameters = {
            SELECTED_SUBJECTS: [""],
            SELECTED_BEHAVIORS: ["p", "s"],
            config.INCLUDE_MODIFIERS: False,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "|"

        out = export_observation.events_to_behavioral_sequences(pj, obs_id, subject, parameters, behav_seq_separator)

        # open("1", "w").write(out)
        assert open("files/Test_events_to_behavioral_sequences_test_5_observation_not_paired").read() == out

    @pytest.mark.usefixtures("before")
    def test_6_multirow_description(self):

        pj = json.loads(open("files/test.boris").read())

        observations = ["live export behavioral sequences"]
        parameters = {
            SELECTED_SUBJECTS: [""],
            SELECTED_BEHAVIORS: ["p", "s"],
            config.INCLUDE_MODIFIERS: False,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "|"

        r, msg = export_observation.observation_to_behavioral_sequences(
            pj,
            selected_observations=observations,
            parameters=parameters,
            behaviors_separator=behav_seq_separator,
            separated_subjects=True,
            timed=False,
            file_name="output/Test_events_to_behavioral_sequences_test_6_multi_rows_description.txt",
        )

        assert (
            open("output/Test_events_to_behavioral_sequences_test_6_multi_rows_description.txt").read()
            == open("files/Test_events_to_behavioral_sequences_test_6_multi_rows_description.txt").read()
        )

    @pytest.mark.usefixtures("before")
    def test_7_all_subjects_2_observations(self):
        """
        all subjects
        2 observations
        2 behaviors
        """

        pj = json.loads(open("files/test.boris").read())

        observations = ["live export behavioral sequences", "observation #1"]
        parameters = {
            SELECTED_SUBJECTS: ["", "subject1", "subject2"],
            SELECTED_BEHAVIORS: ["p", "s"],
            config.INCLUDE_MODIFIERS: False,
            config.EXCLUDE_BEHAVIORS: False,
            config.START_TIME: 0,
            config.END_TIME: 100.0,
        }

        behav_seq_separator = "|"

        r, msg = export_observation.observation_to_behavioral_sequences(
            pj,
            selected_observations=observations,
            parameters=parameters,
            behaviors_separator=behav_seq_separator,
            separated_subjects=True,
            timed=False,
            file_name="output/Test_events_to_behavioral_sequences_test_7.txt",
        )

        assert (
            open("output/Test_events_to_behavioral_sequences_test_7.txt").read()
            == open("files/Test_events_to_behavioral_sequences_test_7.txt").read()
        )


"""
open("1", "w").write(out)

a= Test_export_events()
a.test_export_tabular_xlsx()
"""



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

from PyQt5.QtCore import QPoint, Qt
from PyQt5.QtGui import QPolygon, QPen, QColor, QBrush
from PyQt5.QtWidgets import QStyledItemDelegate


class StyledItemDelegateTriangle(QStyledItemDelegate):
    """
    painter for tv_events with current time highlighting
    """

    def __init__(self, row, parent=None):
        super(StyledItemDelegateTriangle, self).__init__(parent)
        self.row = row

    def paint(self, painter, option, index):
        super(StyledItemDelegateTriangle, self).paint(painter, option, index)

        if self.row == -1:
            return
        if index.row() == self.row:
            polygonTriangle = QPolygon(3)
            polygonTriangle.setPoint(0, QPoint(option.rect.x() + 15, option.rect.y()))
            polygonTriangle.setPoint(1, QPoint(option.rect.x(), option.rect.y() - 5))
            polygonTriangle.setPoint(2, QPoint(option.rect.x(), option.rect.y() + 5))
            painter.save()
            painter.setRenderHint(painter.Antialiasing)
            painter.setBrush(QBrush(QColor(Qt.red)))
            painter.setPen(QPen(QColor(Qt.red)))
            painter.drawPolygon(polygonTriangle)
            painter.restore()



import sys
import time
import pprint
import pandas as pd

from . import project_functions
from . import config as cfg
from . import utilities as util


def event_val(event, field, obs_type: str = cfg.MEDIA):
    try:
        return event[cfg.PJ_OBS_FIELDS[obs_type][field]]
    except:
        return None


_, _, pj, _ = project_functions.open_project_json(sys.argv[1])

# pprint.pprint(list(pj[cfg.OBSERVATIONS].keys()))

# print()

"""
obs_id = "images NO TIME"
pprint.pprint(pj[cfg.OBSERVATIONS][obs_id])
"""

state_events_list = util.state_behavior_codes(pj[cfg.ETHOGRAM])

df_def = {
    "observation id": pd.Series(dtype="str"),
    "observation type": pd.Series(dtype="str"),
    "observation description": pd.Series(dtype="str"),
    "subject": pd.Series(dtype="str"),
    "behavior": pd.Series(dtype="str"),
    "modifier": pd.Series(dtype="str"),
    "start": pd.Series(dtype="float"),
    "stop": pd.Series(dtype="float"),
    "duration": pd.Series(dtype="float"),
}
df = pd.DataFrame(df_def)
l = []
# print(df.info)

for obs_id in pj[cfg.OBSERVATIONS]:
    # print(obs_id)
    obs_type = pj[cfg.OBSERVATIONS][obs_id][cfg.TYPE]
    obs_descr = pj[cfg.OBSERVATIONS][obs_id][cfg.DESCRIPTION]
    mem_idx = []
    for idx, event in enumerate(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]):
        if idx in mem_idx:
            continue
        start = event_val(event, cfg.TIME)
        subject = event_val(event, cfg.SUBJECT)
        behavior = event_val(event, cfg.BEHAVIOR_CODE)
        modifier = event_val(event, cfg.MODIFIER)
        if behavior in state_events_list:
            stop = None
            for idx2, event2 in enumerate(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][idx + 1 :], start=idx + 1):
                subject2 = event_val(event2, cfg.SUBJECT)
                behavior2 = event_val(event2, cfg.BEHAVIOR_CODE)
                modifier2 = event_val(event2, cfg.MODIFIER)
                if subject == subject2 and behavior == behavior2 and modifier == modifier2:
                    stop = event_val(event2, cfg.TIME)
                    mem_idx.append(idx2)
                    l.append([obs_id, obs_type, obs_descr, subject, behavior, modifier, start, stop, stop - start])
                    break
            if stop is None:
                print(obs_id, " not paired")
                # print(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS])
                # print(f"{l=}")
                # print(f"{mem_idx=}")

                sys.exit()
        else:
            l.append([obs_id, obs_type, obs_descr, subject, behavior, modifier, start, start, 0])

# print(pd.DataFrame(l))
df = pd.concat(
    [
        df,
        pd.DataFrame(
            l,
            columns=[
                "observation id",
                "observation type",
                "observation description",
                "subject",
                "behavior",
                "modifier",
                "start",
                "stop",
                "duration",
            ],
        ),
    ]
)
del l
print("=" * 30)
print("describe")
print(df.describe())
print("=" * 30)

# print(f'{df["subject"].value_counts()=}')
# print(f'{df["subject"].nunique()=}')

pd.set_option("display.max_rows", None, "display.max_columns", None)

print("=" * 30)
print("mean")
r = df.groupby(["subject", "behavior"])["duration"].mean()
print(r)
print("=" * 30)


r = df.groupby(["observation id", "subject", "behavior"])
print(r["start"])


"""
# replace value (for selecting a time interval)
t1 = time.time()
df.loc[df["stop"] > 10, "stop"] = 10
print(time.time() - t1)
print(df)
"""


"""
t1 = 1
t2 = 2
print(df.query(f"`start` <= {t1} & `stop` >= {t2} & `duration` != 0"))
"""
